from __future__ import print_function
import sys
import json
from os import environ
from os.path import splitext
from itertools import groupby
import logging.config as LC
import logging

import six
from tqdm import tqdm
import yaml
from pkg_resources import iter_entry_points, DistributionNotFound

from .cli_command_wrapper import CLICommandWrapper, CLIUserError
from .cli_hints import CLI_HINTS
from .command import OWM
from .git_repo import GitRepoProvider
from .text_util import format_table
from .command_util import GeneratorWithData, GenericUserError, SubCommand


CLI_HINTS_GROUP = 'owmeta_core.cli_hints'
COMMANDS_GROUP = 'owmeta_core.commands'

L = logging.getLogger(__name__)


def additional_args(parser):
    'Add some additional options specific to CLI'
    # The "Default is '<blah>'" part of help is to match the cli_command_wrapper output
    parser.add_argument('--output-mode', '-o', default='text',
            help='How to print the results of a command'
            ' (if any). Default is \'text\'',
            choices=['json', 'text', 'table'])
    parser.add_argument('--columns',
            help='Comma-separated list of columns to display in "table" output mode')
    parser.add_argument('--text-pair-separator', default='\t',
            help=r'Separator to use between key and value in "text" output mode. Default is'
            r" '\t' (tab character)")
    parser.add_argument('--text-field-separator', default='\n',
            help=r'Separator to use between fields in "text" output mode. Default is'
            r" '\n' (newline character)")
    parser.add_argument('--text-record-terminator', default='\n',
            help='Terminator to use after each record in "text" output mode. Default is'
            r" '\n' (newline character)")
    parser.add_argument('--text-format',
            help='Format for text output. Available formats determined by the command')
    parser.add_argument('--progress',
            help='Progress reporter to use. Default is \'tqdm\'',
            choices=['tqdm', 'none'],
            default='tqdm')
    parser.add_argument('--full-trace',
            help='Show full stack trace for all uncaught exceptions.',
            action='store_true')
    parser.add_argument('--logging-config',
            help='Set the logging config file')
    parser.add_argument('--log-level',
            help='Set the root log level. Default is \'WARNING\'',
            default='WARNING')


def parse_progress(s):
    if s == 'tqdm':
        return tqdm


def die(message, status=1):
    print(message, file=sys.stderr)
    raise SystemExit(1)


NOT_SET = object()


class NSHandler(object):
    def __init__(self, command, **kwargs):
        self.command = command
        self.opts = dict(kwargs)

    def __getitem__(self, k):
        return self.opts[k]

    def __getattr__(self, k, default=NOT_SET):
        if default is NOT_SET:
            try:
                return self.opts[k]
            except KeyError as e:
                raise AttributeError(k) from e
        else:
            return self.opts.get(k, default)

    def __call__(self, ns):
        self.opts['output_mode'] = ns.output_mode
        self.opts['text_pair_separator'] = ns.text_pair_separator
        self.opts['text_field_separator'] = ns.text_field_separator
        self.opts['text_record_terminator'] = ns.text_record_terminator
        self.opts['text_format'] = ns.text_format
        self.opts['columns'] = ns.columns
        self.opts['full_trace'] = ns.full_trace
        prog = parse_progress(ns.progress)
        if prog:
            self.command.progress_reporter = prog
        if ns.log_level is not None:
            logging.getLogger().setLevel(getattr(logging, ns.log_level.upper()))

        if ns.logging_config:
            cfg_fname = ns.logging_config
            try:
                _, ext = splitext(cfg_fname)
                if ext in ('.yml', '.yaml', '.json'):
                    with open(cfg_fname) as f:
                        LC.dictConfig(yaml.safe_load(f))
                else:
                    LC.fileConfig(ns.logging_config)
            except Exception:
                print('Unable to load logging configuration from a file')
                raise
        else:
            logging.basicConfig()

    def __str__(self):
        return 'NSHandler' + str(self.opts)


class JSONSerializer(object):
    def __call__(self, o):
        from rdflib.graph import Graph
        from owmeta_core.context import Context
        if isinstance(o, Graph):
            # eventually, we will use something like JSON-LD
            return []
        elif isinstance(o, Context):
            return {'identifier': o.identifier,
                    'base_namespace': o.base_namespace}
        else:
            return list(o)


def _select(a, indexes):
    return [h for h, i in zip(a, range(len(a))) if i in indexes]


def columns_arg_to_list(arg):
    return [s.strip() for s in arg.split(',')]


def main(*args):
    '''
    Entry point for the command line interface.

    Additional sub-commands can be added by specifying them in an entry point in your
    package's setup.py like this::

        'owmeta_core.commands': [
            'subcommand_name = module.path.for:TheSubCommand',
            'sub.sub.subcommand_name = module.path.for:TheSubSubSubCommand',
        ],

    Where, ``subcommand_name`` will be the name of the sub-command under the top-level
    ``owm`` command and ``module.path.for.TheSubCommand`` will be the class implementing
    the command. To add to existing sub-commands one indicate the place in the command
    hierarchy as in ``sub.sub.subcommand_name``: ``TheSubSubSubCommand`` would be
    available under the (hypothetical) existing ``owm sub sub`` command as ``owm sub sub
    subcommand_name``

    So-called "hints" can affect the way command implementations are interepreted such as
    indicating whether a method argument should be read in as a positional argument or
    an option and what a command-line option should be named (as opposed to deriving it
    from a parameter name or member variable). There is a set of hints which are a part of
    owmeta-core (see `CLI_HINTS`), but these can be augmented by specifying entry points
    like this::

        'owmeta_core.cli_hints': 'hints = module.path.for:CLI_HINTS',

    If ``module.path.for.CLI_HINTS`` is a dictionary, it will get added to the hints,
    potentially affecting any sub-commands without hints already available. The entry
    point name (``hints`` in the example) is only used for error-reporting by this module.
    Although this is not strictly enforced, adding hints for sub-commands published by
    other modules, including owmeta-core, should be avoided to ensure consistent behavior
    across installations. See `owmeta_core.cli_hints` source for the format of hints.

    See `CLICommandWrapper` for more details on how the command line options are constructed.

    Parameters
    ----------
    *args
        Arguments to the command. Used instead of `sys.argv`
    '''
    top_command = _augment_subcommands_from_entry_points()
    p = top_command()
    p.message = print
    p.repository_provider = GitRepoProvider()
    p.non_interactive = False
    if environ.get('OWM_CLI_PROFILE'):
        from cProfile import Profile
        profiler = Profile()
        profiler.enable()

    try:
        _helper(p, args=(args or None))
    except (CLIUserError, GenericUserError) as e:
        s = str(e)
        if not s:
            from .utils import FCN
            # In case someone forgets to add a helpful message for their user error
            s = 'Received error: ' + FCN(type(e))
        die(s)

    if environ.get('OWM_CLI_PROFILE'):
        profiler.disable()
        profiler.dump_stats(environ['OWM_CLI_PROFILE'])


def _gather_hints_from_entry_points(existing_hints):
    res = existing_hints
    for entry_point in iter_entry_points(group=CLI_HINTS_GROUP):
        try:
            for cmd, hints in entry_point.load().items():
                if cmd in res:
                    L.warning(f'Hints are already defined for {cmd}: {entry_point}'
                            f' hints for {cmd} will be ignored')
                    continue
                res[cmd] = hints

        except Exception:
            L.warning('Unable to add CLI hints for %s', entry_point, exc_info=True)
    return res


def _augment_subcommands_from_entry_points():
    command_map = {(): OWM}
    unordered_commands = []
    for entry_point in iter_entry_points(COMMANDS_GROUP):
        name = str(entry_point.name).strip()
        path = tuple(name.split('.'))
        try:
            unordered_commands.append((path, entry_point.load()))
        except DistributionNotFound:
            # This is expected...there's no pre-filtering of entry points for when extras
            # aren't installed...maybe to allow for optional depend on extras
            L.debug('Not adding sub-command %s due to failure in package resources resolution',
                    entry_point, exc_info=True)
        except Exception:
            L.warning('Unable to add command %s', entry_point, exc_info=True)
    level_ordered_commands = \
        sorted(unordered_commands, key=lambda x: (len(x[0]), x[0]))
    # Group by everything up-to a given command
    for indicator, level in groupby(level_ordered_commands,
            key=lambda x: (len(x[0]), x[0][:-1])):
        override_dict = dict()
        level_key = None
        for path, cmd in level:
            level_key = path[:-1]
            if isinstance(cmd, type):
                override_dict[path[-1]] = SubCommand(cmd)
            else:
                override_dict[path[-1]] = cmd

        if level_key is not None:
            orig_cmd = command_map.get(level_key)
            if not orig_cmd:
                last = command_map[()]
                # Get the commands indicated by the prefixes of the key
                for idx, component in enumerate(level_key):
                    cmd = command_map.get(level_key[:idx+1])
                    if not cmd:
                        # Get the SubCommand object matching the current end of the
                        # prefix, get the class it contains and add it to the command_map
                        cmd = getattr(last, component).cmd
                        command_map[level_key[:idx+1]] = cmd
                    last = cmd
                orig_cmd = last

            # Subclass the parent command
            override_dict['__module__'] = orig_cmd.__module__
            command_map[level_key] = type(orig_cmd.__name__, (orig_cmd,), override_dict)
            if level_key:
                setattr(command_map[level_key[:-1]], level_key[-1], SubCommand(command_map[level_key]))

    return command_map[()]


def _helper(p, args=None):
    ns_handler = NSHandler(p)

    hints = _gather_hints_from_entry_points(dict(**CLI_HINTS))

    try:
        out = CLICommandWrapper(p, hints_map=hints).main(
                args=args,
                argument_callback=additional_args,
                argument_namespace_callback=ns_handler)

        if out is not None:
            _format_output(out, ns_handler)
    except (GenericUserError, CLIUserError):
        try:
            if ns_handler.full_trace:
                import traceback
                traceback.print_exc()
        finally:
            raise


def _format_output(out, ns_handler):
    output_mode = ns_handler.output_mode
    text_pair_separator = ns_handler.text_pair_separator
    text_field_separator = ns_handler.text_field_separator
    text_record_terminator = ns_handler.text_record_terminator
    text_format = ns_handler.text_format

    if output_mode == 'json':
        json.dump(out, sys.stdout, default=JSONSerializer(), indent=2)
    elif output_mode == 'table':
        # `out.header` holds the *names* for each column whereas `out.columns` holds
        # accessors for each column. `out` must have either:
        # 1) Just `header`
        # 2) `header` and `columns`
        # 3) neither `header` nor `columns`
        # at least one to be valid.
        #
        # If `out` has only `header`, then `header must have one element which is the
        # header for a single result from `out`.
        #
        # If it has `header` and `columns` then the two must be of equal length and
        # each element of `header` names the result of the corresponding entry in
        # `columns`.
        #
        # If neither `header` nor `columns` is given, then a default header of
        # `['Value']` is used and the column accessor is the identify function.
        #
        # `out.default_columns` determines which columns to show by default. If there
        # is no `default_columns`, then all defined columns are shown. The columns can
        # be overriden by the `--columns` command line argument. If there is no
        # `header`, but `--columns` is provided, the user will get an error.

        # N.B.: we use getattr rather than hasattr because we want to make sure we
        # have the attribute AND that the columns/header isn't an empty or tuple/list
        if getattr(out, 'columns', None) and getattr(out, 'header', None):
            if ns_handler.columns:
                cols = columns_arg_to_list(ns_handler.columns)
                selected_columns = [i for i, e in zip(range(len(out.header)), out.header)
                                    if e in cols]
                if not selected_columns or len(selected_columns) != len(cols):
                    die('The given list of columns is not valid for this command')
            elif getattr(out, 'default_columns', None):
                selected_columns = [i for i, e in zip(range(len(out.header)), out.header)
                                    if e in out.default_columns]
            else:
                selected_columns = list(range(len(out.header)))
        elif getattr(out, 'columns', None):
            raise Exception('Programmer error: A header must be provided if an output'
                            ' has multiple columns')
        elif getattr(out, 'header', None):
            if len(out.header) != 1:
                raise Exception('Programmer error: Only one column in the header can be defined if'
                                ' no columns are defined')
            if ns_handler.columns \
                    and columns_arg_to_list(ns_handler.columns) != out.header:
                die('The given list of columns is not valid for this command')
            out = GeneratorWithData(out,
                                    columns=[lambda x: x],
                                    header=out.header)
            selected_columns = [0]
        elif ns_handler.columns:
            die('The given list of columns is not valid for this command')
        else:
            out = GeneratorWithData(out,
                                    columns=[lambda x: x],
                                    header=['Value'])
            selected_columns = [0]

        header = _select(out.header, selected_columns)
        columns = _select(out.columns, selected_columns)

        def mygen():
            if columns:
                for m in out:
                    yield tuple(c(m) for c in columns)
            else:
                for m in out:
                    yield (m,)
        print(format_table(mygen(), header=header))
    elif output_mode == 'text':
        if isinstance(out, dict):
            first = True
            for k, v in out.items():
                if not first:
                    print(text_field_separator, end='')
                first = False
                print('{}{}{}'.format(k, text_pair_separator, v), end='')
            print(text_record_terminator, end='')
        elif isinstance(out, six.string_types):
            print(out)
        else:
            try:
                iterable = (x for x in out)
            except TypeError:
                print(out)
            else:
                for x in iterable:
                    if getattr(out, 'text_format', None):
                        addl_args = dict()
                        if text_format:
                            addl_args['format'] = text_format

                        print(out.text_format(x, **addl_args), end='')
                    elif isinstance(x, dict):
                        first = True
                        for k, v in x.items():
                            if not first:
                                print(text_field_separator, end='')
                            first = False
                            print('{}{}{}'.format(k, text_pair_separator, v), end='')
                    else:
                        print(x, end='')
                    print(text_record_terminator, end='')


if __name__ == '__main__':
    main()



from __future__ import print_function

import itertools
import logging
from importlib import import_module

import rdflib as R
from six import with_metaclass

from . import RDF_CONTEXT, BASE_SCHEMA_URL, BASE_DATA_URL
from .utils import FCN
from .data import DataUser
from .context import Context
from .contextualize import (Contextualizable, ContextualizableClass,
                            contextualize_helper,
                            decontextualize_helper)
from .context_mapped_class_util import find_class_context, find_base_namespace
from .graph_object import GraphObject, GraphObjectQuerier
from .inverse_property import InversePropertyMixin
from .mapped_class import MappedClass
from .property_mixins import UnionPropertyMixin
from .property_value import PropertyValue
from .rdf_utils import deserialize_rdflib_term
from .rdf_query_modifiers import (rdfs_subpropertyof_zom,
                                  rdfs_subclassof_zom,
                                  ZeroOrMoreTQLayer,
                                  ContainerMembershipIsMemberTQLayer)
from .rdf_query_util import goq_hop_scorer, load_base
from .statement import Statement
from .variable import Variable

L = logging.getLogger(__name__)


class ContextMappedPropertyClass(MappedClass, ContextualizableClass):
    '''
    Meta-class for `Property`.

    A few attributes can be specified in the class body which affect how the created type
    is set up: these are defined in the "Attributes" section.

    One aspect in particular is important: a `Property` class can represent a single type
    of property where all instances have the same URI, or a `Property` can represent an a
    class of RDF properties where the instances have distinct URIs and correspond to
    instances of the RDF type. An instance of the latter is demonstrated with
    `~owmeta_core.collections.ContainerMembershipProperty`.

    Attributes
    ----------
    rdf_type_class : type
        A sub-class of `~.dataobject.DataObject` to use as the type. If set,
        this will be used instead of what `init_rdf_type_object` would create.
    rdf_type : str or URIRef
        The RDF type for the `Property`. Must be defined for `init_rdf_type_object` to
        actually create the rdf type object
    rdf_type_object_deferred : bool
        If `True`, defer calling `init_rdf_type_object` until it's explicitly called rather
        than during normal class init. Useful for cases where `init_rdf_type_object` uses
        types that aren't defined at the point where the `Property` is defined.
    rdf_object : RDFProperty
        An instance of `RDFProperty` corresponding to this class. If set, this will be
        used instead of what `init_rdf_object` would create.
    rdf_object_deferred : bool
        If `True`, defer calling `init_rdf_object` until it is explicitly called rather
        than during normal class init. Useful for cases where `init_rdf_object` uses types
        that aren't defined at the point where the `Property` is defined.

    '''

    context_carries = ('rdf_object_deferred', 'rdf_type_object_deferred', 'link', 'linkName', 'rdf_type_class')

    def __init__(self, name, bases, dct):
        super(ContextMappedPropertyClass, self).__init__(name, bases, dct)
        ctx = find_class_context(self, dct, bases)

        self.linkName = dct.get('link_name', dct.get('linkName'))
        self.link_name = self.linkName

        self.link = dct.get('link')

        if self.link is None and self.linkName is not None:
            self.link = self.base_namespace[self.linkName]

        if getattr(self, 'value_type', None) is not None and getattr(self, 'value_rdf_type', None) is None:
            self.value_rdf_type = self.value_type.rdf_type

        if 'definition_context' in dct:
            self.__definition_context = dct['definition_context']
        elif ctx is not None:
            self.__definition_context = ctx
        else:
            self.__definition_context = None

        if not hasattr(self, 'base_namespace') or self.base_namespace is None:
            self.base_namespace = find_base_namespace(dct, bases)

        self.rdf_type_class = dct.get('rdf_type_class')
        self.__rdf_object = dct.get('rdf_object')
        self.__rdf_object_callback = dct.get('rdf_object_callback')

        self.rdf_object_deferred = dct.get('rdf_object_deferred', False)
        self.rdf_type_object_deferred = dct.get('rdf_type_object_deferred', False)

        # We have to deferr initializing our rdf_object since initialization depends on
        # RDFSSubClassOfProperty being initialized
        self.subproperty_of = dct.get('subproperty_of')
        if not self.rdf_object_deferred:
            self.init_rdf_object()

        if not self.rdf_type_object_deferred:
            self.init_rdf_type_object()

        if not getattr(self, 'unmapped', False) and not dct.get('unmapped'):
            module = import_module(self.__module__)
            if not hasattr(module, '__yarom_mapped_classes__'):
                module.__yarom_mapped_classes__ = [self]
            else:
                module.__yarom_mapped_classes__.append(self)

    @property
    def rdf_object(self):
        if self.__rdf_object_callback is not None:
            rdto = self.__rdf_object_callback()
            if rdto is not None:
                self.__rdf_object_callback = None
                self.__rdf_object = rdto
        return self.__rdf_object

    @rdf_object.setter
    def rdf_object(self, value):
        if value is not None:
            self.__rdf_object_callback = None
        self.__rdf_object = value

    def contextualize_class_augment(self, context):
        '''
        For MappedClass, rdf_type and rdf_namespace have special behavior where they can
        be auto-generated based on the class name and base_namespace. We have to pass
        through these values to our "proxy" to avoid this behavior
        '''
        args = dict()
        if self.rdf_object is None:
            args['rdf_object_callback'] = lambda: self.rdf_object
        else:
            args['rdf_object'] = self.rdf_object

        res = super(ContextMappedPropertyClass, self).contextualize_class_augment(
                context,
                **args)
        res.__module__ = self.__module__
        return res

    def init_rdf_object(self):
        if (self.link is not None and
                (self.rdf_object is None or
                    self.rdf_object.identifier != self.link)):
            from .dataobject import RDFProperty
            if self.definition_context is None:
                L.info("The class %s has no context for RDFProperty(ident=%s)",
                    self, self.link)
                return
            L.debug('Creating rdf_object for %s in %s', self, self.definition_context)
            rdto = RDFProperty.contextualize(self.definition_context)(ident=self.link)
            if hasattr(self, 'label'):
                rdto.rdfs_label(self.label)

            superproperty = getattr(self, 'subproperty_of', None)
            if superproperty is not None:
                if isinstance(superproperty, (list, tuple)):
                    for sp in superproperty:
                        rdto.rdfs_subpropertyof(sp.rdf_object)
                else:
                    rdto.rdfs_subpropertyof(superproperty.rdf_object)

            self.rdf_object = rdto

    def init_rdf_type_object(self):
        '''
        Initializes `rdf_type_class` and thereby initializes the `rdf_type_object`

        Sometimes, we actually use Property sub-classes *as* ``rdf:Property`` classes
        (e.g., ``rdfs:ContainerMembershipProperty``). The `rdf_type` attribute has to be
        defined on this class if we're going to use it as an ``rdf:Property`` class.
        '''
        from .dataobject import RDFProperty

        rdf_type = getattr(self, 'rdf_type', None)
        if (rdf_type is not None and
                getattr(self, 'rdf_object', None) is None and
                getattr(self, 'rdf_type_class', None) is None):
            self.rdf_type_class = type(self.__name__,
                    (RDFProperty,),
                    dict(rdf_type=rdf_type,
                         class_context=self.definition_context))

    @property
    def definition_context(self):
        return self.__definition_context

    def declare_class_registry_entry(self):
        from owmeta_core.dataobject import RegistryEntry
        re = RegistryEntry.contextualize(self.context)()
        cd = self.declare_class_description()

        re.rdf_class(self.rdf_type)
        re.class_description(cd)

    def declare_class_description(self):
        from owmeta_core.dataobject import PythonClassDescription, PythonModule
        cd = PythonClassDescription.contextualize(self.context)()

        mo = PythonModule.contextualize(self.context)()
        mo.name(self.__module__)

        cd.module(mo)
        cd.name(self.__name__)
        return cd

    def __lt__(self, other):
        res = False
        if issubclass(other, self) and not issubclass(self, other):
            res = True
        elif issubclass(self, other) == issubclass(other, self):
            res = self.__name__ < other.__name__
        return res


class ContextualizedPropertyValue(PropertyValue):

    @property
    def context(self):
        return None


class _StatementContextRDFObjectFactory(Contextualizable):
    __slots__ = ('context', 'statement')

    def __init__(self, statement):
        self.context = None
        self.statement = statement

    def contextualize(self, context):
        temp = _StatementContextRDFObjectFactory(self.statement)
        temp.context = context
        return temp

    def __call__(self):
        if self.context is None:
            raise ValueError("No context has been set for this proxy")
        return self.statement.context.contextualize(self.context).rdf_object

    def __repr__(self):
        return '{}({})'.format(FCN(type(self)), repr(self.statement))


class Property(with_metaclass(ContextMappedPropertyClass, DataUser, Contextualizable)):
    '''
    A property attached to a `~owmeta_core.dataobject.DataObject`.
    '''

    multiple = False
    '''
    If `True`, then the property will only maintain a single staged value at a time. No
    effort is made to check how many values are stored in the RDF graph.
    '''

    class_context = RDF_CONTEXT
    link = None
    linkName = None
    cascade_retract = False
    base_namespace = R.Namespace(BASE_SCHEMA_URL + '/')
    base_data_namespace = R.Namespace(BASE_DATA_URL + '/')

    lazy = True
    '''
    If `True`, then the property is not attached to an instance until the property is set
    or queried.
    '''

    rdf_object_deferred = True
    rdf_type_object_deferred = True

    def __init__(self, owner, **kwargs):
        super(Property, self).__init__(**kwargs)
        self._v = []
        self.owner = owner
        self._hdf = dict()
        self._expr = None

    @property
    def expr(self):
        '''
        An query expression from this property
        '''
        if self._expr is None:
            self._expr = PropertyExpr([self])
        return self._expr

    def contextualize_augment(self, context):
        self._hdf[context] = None
        res = contextualize_helper(context, self)
        if res is not self:
            cowner = context(res.owner)
            res.add_attr_override('owner', cowner)
        return res

    def decontextualize(self):
        self._hdf[self.context] = None
        return decontextualize_helper(self)

    def has_value(self):
        '''
        Returns `True` if there is a value set on this property in the currrent context
        '''
        for x in self._v:
            if x.context == self.context:
                return True
        return False

    def has_defined_value(self):
        '''
        Returns `True` if this property has a value in the current context which is either
        a `GraphObject` with `defined` set to `True` or a literal value
        '''
        hdf = self._hdf.get(self.context)
        if hdf is not None:
            return hdf
        for x in self._v:
            if x.context == self.context and x.object.defined:
                self._hdf[self.context] = True
                return True
        return False

    def set(self, v):
        '''
        Set the value for or add a value to this property
        '''
        if v is None:
            raise ValueError('It is not permitted to declare a property to have value the None')

        if not hasattr(v, 'idl'):
            v = ContextualizedPropertyValue(v)

        if not self.multiple:
            self.clear()

        stmt = self._insert_value(v)
        if self.context is not None:
            self.context.add_statement(stmt)
        return stmt

    def clear(self):
        """ Clears values set *in all contexts* """
        self._hdf = dict()
        for x in self._v:
            assert self in x.object.owner_properties
            x.object.owner_properties.remove(self)
            self._v.remove(x)

    @property
    def defined_values(self):
        '''
        The "defined" values set on this property in the current context
        '''
        return tuple(x.object for x in self._v
                     if x.object.defined and x.context == self.context)

    @property
    def values(self):
        '''
        Return all values set on this property in the current context
        '''
        return tuple(self._values_helper())

    def _values_helper(self):
        context = self.context
        for x in self._v:
            if x.context == context:
                # XXX: decontextualzing default context here??
                if context is not None:
                    yield context(x.object)
                elif isinstance(x.object, Contextualizable):
                    yield x.object.decontextualize()
                else:
                    yield x.object

    @property
    def rdf(self):
        if self.context is not None:
            return self.context.rdf_graph()
        else:
            return super(Property, self).rdf

    @property
    def identifier(self):
        '''
        Alias to `link`
        '''
        return self.link

    def get(self):
        if self.rdf is None:
            return ()
        results = None
        owner = self.owner

        g = ZeroOrMoreTQLayer(rdfs_subclassof_zom, self.rdf)
        g = ContainerMembershipIsMemberTQLayer(g)
        g = ZeroOrMoreTQLayer(rdfs_subpropertyof_zom(self.link), g)
        if owner.defined:
            results = set()
            ident = owner.identifier
            for s, p, o in g.triples((ident, self.link, None)):
                results.add(o)
        else:
            v = Variable("var" + str(id(self)))
            self._insert_value(v)
            results = GraphObjectQuerier(v, g, hop_scorer=goq_hop_scorer)()
            self._remove_value(v)
        return results

    get_terms = get
    '''
    Get the `~rdflib.term.Node` instances matching this property query
    '''

    def _insert_value(self, v):
        stmt = Statement(self.owner, self, v, self.context)
        self._hdf[self.context] = None
        self._v.append(stmt)
        v.owner_properties.append(self)
        return stmt

    def _remove_value(self, v):
        assert self in v.owner_properties
        self._hdf[self.context] = None
        v.owner_properties.remove(self)
        self._v.remove(Statement(self.owner, self, v, self.context))

    unset = _remove_value
    '''
    Remove a from this property
    '''

    def __call__(self, *args, **kwargs):
        """ If arguments are given ``set`` method is called. Otherwise, the ``get``
        method is called. If the ``multiple`` member is set to ``True``, then a
        Python set containing the associated values is returned. Otherwise, a
        single bare value is returned.
        """
        if len(args) > 0 or len(kwargs) > 0:
            return self.set(*args, **kwargs)
        else:
            r = self.get(*args, **kwargs)
            if self.multiple:
                return set(r)
            else:
                for a in r:
                    return a
                return None

    def __repr__(self):
        fcn = FCN(type(self))
        return '{}(owner={})'.format(fcn, repr(getattr(self, 'owner', None)))

    def one(self):
        '''
        Query for a single value from this property.

        For a multi-valued property, the returned value is chosen arbitrarily. If there's
        no value returned from the query, then `None` is returned.

        '''
        return next(iter(self.get()), None)

    def onedef(self):
        '''
        Return a single defined value set on this property in the current context

        This does not execute a query, but returns a value which was set on this property.
        '''
        for x in self._v:
            if x.object.defined and x.context == self.context:
                return x.object
        return None

    @property
    def defined_statements(self):
        return tuple(x for x in self._v
                     if x.object.defined and x.subject.defined)

    @property
    def statements(self):
        return self.rdf.quads((self.owner.idl, self.link, None, None))


class PropertyExpr(object):
    '''
    A property expression
    '''
    def __init__(self, props, triples_provider=None, terms_provider=None, origin=None):
        self.props = props
        self.prop = props[0]
        if origin is None:
            origin = self.prop
        self.origin = origin
        self.rdf = self.origin.rdf
        self.terms = None
        if triples_provider is None:
            triples_provider = self._make_triples

        self.triples_provider = triples_provider

        if terms_provider is None:
            terms_provider = self._compute_terms

        self.terms_provider = terms_provider

        self.created_sub_expressions = dict()
        self.dict = None
        self.combos = []

    def __or__(self, other):
        if self is other:
            return self

        def terms_provider():
            return itertools.chain(self.terms_provider(), other.terms_provider())

        def triples_provider():
            return itertools.chain(self.triples_provider(), other.triples_provider())

        res = type(self)(self.props + other.props,
                terms_provider=terms_provider,
                triples_provider=triples_provider,
                origin=self.origin)

        self.combos.append(res)
        other.combos.append(res)

        return res

    @property
    def rdf_type(self):
        '''
        Short-hand for `rdf_type_property`
        '''
        return self.rdf_type_property

    # Careful not to define any property descriptors below this definition
    def property(self, property_class):
        '''
        Create a sub-expression with the given property.

        Allows for creating expressions with properties that are not necessarily declared
        for the `value_type` of this expression's property
        '''
        if ('link', property_class.link) in self.created_sub_expressions:
            return self.created_sub_expressions[('link', property_class.link)]
        res = self._create_sub_expression(property_class)
        self.created_sub_expressions[('link', property_class.link)] = res
        return res

    def __getattr__(self, attr):
        if ('attr', attr) in self.created_sub_expressions:
            return self.created_sub_expressions[('attr', attr)]
        value_types = []
        for prop in self.props:
            try:
                value_types.append(getattr(prop, 'value_type'))
            except AttributeError:
                raise AttributeError(attr)

        sub_props = []
        for v in value_types:
            sub_prop = getattr(v, attr)
            sub_props.append(sub_prop)

        res = None
        for sub_prop in sub_props:
            new_expr = self._create_sub_expression(sub_prop)
            if res is None:
                res = new_expr
                continue
            res = res | new_expr
        self.created_sub_expressions[('attr', attr)] = res
        return res

    def _create_sub_expression(self, sub_prop):
        link = sub_prop.link
        triples_choices = self.rdf.triples_choices

        def terms_provider():
            terms = list(self.terms_provider())
            for c in triples_choices(
                    (terms, link, None)):
                yield c[2]

        def triples_provider():
            terms = list(self.terms_provider())
            for c in triples_choices(
                    (terms, link, None)):
                yield c

        return type(self)([sub_prop],
                terms_provider=terms_provider,
                triples_provider=triples_provider,
                origin=self.origin)

    def to_dict(self, multiple=False):
        '''
        Return a `dict` mapping from identifiers for subjects of this expression's
        property to the objects for that property.

        Parameters
        ----------
        multiple : bool, optional
            If `False`, then only a single object is allowed for each subject in the
            results. An exception is raised if more than one object is found for a given
            subject.
        '''
        if self.dict is None:
            res = dict()
            triples = self.triples_provider()
            if not multiple:
                for s, p, o in triples:
                    current_value = res.get(s)
                    if current_value is not None and current_value != o:
                        raise PropertyExprError(f'More than one value for {p} in the'
                                f' results for {s}')
                    res[s] = o
            else:
                for s, p, o in triples:
                    values = res.get(s)
                    if values is None:
                        values = set([o])
                        res[s] = values
                    else:
                        values.add(o)
            self.dict = res
        return self.dict

    def to_objects(self):
        '''
        Returns a list of `ExprResultObj` that allow for retrieving results in a
        convenient attribute traversal
        '''
        return list(ExprResultObj(self, t) for t in self.to_terms())

    __call__ = to_dict

    def to_terms(self):
        '''
        Return a list of `rdflib.term.Node` terms produced by this expression.
        '''
        if self.terms is None:
            terms = self.terms_provider()
            if not isinstance(terms, list):
                terms = list(terms)
            self.terms = terms
        return self.terms

    def __iter__(self):
        return iter(self.to_terms())

    def __repr__(self):
        return '({}).expr'.format(
                ' | '.join(repr(p) for p in self.props))

    def _make_triples(self):
        # Note: This method only applies for the root property in the expression.
        # Sub-expressions have property *classes* which do not have a 'get_terms' method
        for prop in self.props:
            for term in prop.get_terms():
                yield (prop.owner.identifier, prop.link, term)

    def _compute_terms(self):
        # Note: This method only applies for the root property in the expression.
        # Sub-expressions have property *classes* which do not have a 'get_terms' method
        return list(itertools.chain(*(prop.get_terms() for prop in self.props)))


class PropertyExprError(Exception):
    pass


class ExprResultObj(object):
    '''
    Object returned by `PropertyExpr.to_objects`. Attributes for which
    `PropertyExpr.to_dict` has been called can be accessed on the object. For example we
    can print out the ``b`` properties of instances of a class ``A``::

        class B(DataObject):
            v = DatatypeProperty()

        class A(DataObject):
            b = ObjectProperty(value_type=B)

        a = A().a.expr
        a.b.v()
        for anA in a.to_objects():
            print(anA.identifier, anA.b)

    ``anA`` is an `ExprResultObj` in the example. The
    '''
    __slots__ = ('_expr', 'identifier')

    def __init__(self, expr, ident):
        self._expr = expr
        self.identifier = ident

    @property
    def rdf_type(self):
        '''
        Allias to rdf_type_property
        '''
        return self.rdf_type_property

    def property(self, property_class):
        '''
        Return the results object for this sub-expression

        Parameters
        ----------
        property_class : Property, Property sub-class, URIRef, or str
        '''
        if isinstance(property_class, str):
            link = R.URIRef(property_class)
        else:
            try:
                link = property_class.link
            except AttributeError:
                raise ValueError('Expected either an object with a `link` attribute or a'
                        ' rdflib.term.URIRef or a str')
        sub_expr = self._expr.created_sub_expressions.get(('link', link))
        if not sub_expr:
            for c in self._expr.combos:
                sub_expr = c.created_sub_expressions.get(('link', link))
                if sub_expr:
                    break

        if not sub_expr:
            raise KeyError(property_class)
        return self._give_value(sub_expr)

    def _give_value(self, sub_expr):
        sub_expr_dict = sub_expr.to_dict()
        val = sub_expr_dict.get(self.identifier)
        if val and (sub_expr.created_sub_expressions or
                any(c.created_sub_expressions for c in sub_expr.combos)):
            return type(self)(sub_expr, val)
        else:
            if isinstance(val, R.Literal):
                return deserialize_rdflib_term(val)
            return val

    def __getattr__(self, attr):
        sub_expr = self._expr.created_sub_expressions.get(('attr', attr))
        if not sub_expr:
            for c in self._expr.combos:
                sub_expr = c.created_sub_expressions.get(('attr', attr))
                if sub_expr:
                    break
        if not sub_expr:
            raise AttributeError(attr)
        return self._give_value(sub_expr)

    def __repr__(self):
        return f'{FCN(type(self))}({repr(self._expr)}, {repr(self.identifier)})'


class PropertyCountMixin(object):
    def count(self):
        return sum(1 for _ in super(PropertyCountMixin, self).get())


class ObjectProperty(InversePropertyMixin,
                     PropertyCountMixin,
                     Property):

    rdf_object_deferred = True
    rdf_type_object_deferred = True

    def __init__(self, resolver, *args, **kwargs):
        super(ObjectProperty, self).__init__(*args, **kwargs)
        self.resolver = resolver

    def contextualize_augment(self, context):
        res = super(ObjectProperty, self).contextualize_augment(context)
        if context is not None and self is not res:
            res.add_attr_override('resolver', context(self.resolver))
        return res

    def set(self, v):
        if not isinstance(v, GraphObject):
            raise Exception(
                f"ObjectProperty {self!r} only accepts GraphObject instances. Got {v!r}, a"
                f" {type(v)} a.k.a. " +
                " or ".join(str(x) for x in type(v).__bases__))
        return super(ObjectProperty, self).set(v)

    def get(self):
        idents = super(ObjectProperty, self).get()
        r = load_base(self.rdf,
                      idents,
                      self.value_rdf_type,
                      self.context,
                      self.resolver)
        return itertools.chain(self.defined_values, r)

    @property
    def statements(self):
        for x in self.get():
            yield Statement(self.owner,
                    self,
                    x,
                    self.context)


class DatatypeProperty(PropertyCountMixin, Property):

    rdf_object_deferred = True
    rdf_type_object_deferred = True

    def __init__(self, resolver, **kwargs):
        """
        Parameters
        ----------
        resolver : RDFTypeResolver
            Resolves RDF identifiers returned from :meth:`get` into objects
        """
        super(DatatypeProperty, self).__init__(**kwargs)
        self.resolver = resolver

    def __get(self):
        for v in super(DatatypeProperty, self).get():
            yield self.resolver.deserializer(v)

    def get(self):
        s = set()
        unhashables = []
        for x in self.defined_values:
            val = self.resolver.deserializer(x.idl)
            try:
                s.add(val)
            except TypeError as e:
                unhashables.append(val)
                L.info('Unhashable type: %s', e)
        return itertools.chain(self.__get(), s, unhashables)

    def onedef(self):
        x = super(DatatypeProperty, self).onedef()
        return self.resolver.deserializer(x.identifier) if x is not None else x

    @property
    def statements(self):
        return itertools.chain(self.defined_statements,
                               (Statement(self.owner,
                                          self,
                                          ContextualizedPropertyValue(x[2]),
                                          Context(ident=x[3]))
                                for x in super(DatatypeProperty, self).statements))


class UnionProperty(InversePropertyMixin,
                    UnionPropertyMixin,
                    PropertyCountMixin,
                    Property):

    rdf_object_deferred = True
    rdf_type_object_deferred = True

    """ A Property that can handle either DataObjects or basic types """
    def get(self):
        r = super(UnionProperty, self).get()
        s = set()
        for x in self.defined_values:
            if isinstance(x, PropertyValue):
                s.add(self.resolver.deserializer(x.idl))
            else:
                s.add(x)
        return itertools.chain(r, s)

    def onedef(self):
        x = super(UnionProperty, self).onedef()
        if isinstance(x, PropertyValue):
            return self.resolver.deserializer(x.identifier) if x is not None else x
        return x

    @property
    def statements(self):
        for x in self.get():
            yield Statement(self.owner,
                    self,
                    x,
                    self.context)


def _property_to_string(self):
    try:
        s = str(self.linkName) + "=`" + \
            ";".join(str(s) for s in self.defined_values) + "'"
    except AttributeError:
        s = str(self.linkName) + '(no defined_values)'
    return s



class NotABundlePath(Exception):
    '''
    Thrown when a given path does not point to a valid bundle directory tree or bundle
    archive
    '''
    def __init__(self, path, explanation):
        message = '"{}" is not a bundle path: {}'.format(path, explanation)
        super(NotABundlePath, self).__init__(message)
        self.path = path


class MalformedBundle(NotABundlePath):
    '''
    Thrown when a given path does points to a bundle directory or archive is malformed
    '''


class BundleNotFound(Exception):
    '''
    Thrown when a bundle cannot be found on a local or remote resource with the given
    parameters.
    '''
    def __init__(self, bundle_id, msg=None, version=None):
        '''
        Parameters
        ----------
        bundle_id : str
            ID of the bundle that was sought
        msg : str, optional
            An explanation of why the bundle could not be found
        version : int, optional
            Version number of the bundle
        '''
        msg = 'Missing bundle "{}"{}{}'.format(bundle_id,
                '' if version is None else ' at version ' + str(version),
                ': ' + str(msg) if msg is not None else '')
        super(BundleNotFound, self).__init__(msg)


class InstallFailed(Exception):
    '''
    Thrown when a bundle installation fails to complete.

    You can assume that any intermediate bundle files have been cleaned up from the bundle
    cache
    '''


class UncoveredImports(InstallFailed):
    '''
    Thrown when a bundle to be installed has declared imports but is missing dependencies
    to cover those imports
    '''
    def __init__(self, imports):
        '''
        Parameters
        ----------
        imports : list of URIRef
            List of imports declared for a bundle which are not covered by any of the
            bundle's dependencies
        '''
        msg = 'Missing {} imports'.format(len(imports))
        super(UncoveredImports, self).__init__(msg)
        self.imports = imports


class TargetIsNotEmpty(InstallFailed):
    '''
    Thrown when the target directory of an installation is not empty
    '''
    def __init__(self, target):
        msg = 'Bundle installation target directory, "%s", is not empty' % target
        super(TargetIsNotEmpty, self).__init__(msg)
        self.directory = target


class FetchFailed(Exception):
    ''' Generic message for when a fetch fails '''


class FetchTargetIsNotEmpty(FetchFailed):
    '''
    Thrown when the target directory of a fetch is not empty
    '''
    def __init__(self, target):
        msg = 'Bundle fetch target directory, "%s", is not empty' % target
        super(FetchTargetIsNotEmpty, self).__init__(msg)
        self.directory = target


class NoBundleLoader(FetchFailed):
    '''
    Thrown when a loader can't be found for a bundle
    '''

    def __init__(self, bundle_id, bundle_version=None, message=None):
        super(NoBundleLoader, self).__init__(
            'No loader could be found for "%s"%s%s' % (bundle_id,
                f' at version {bundle_version}' if bundle_version is not None else '',
                f': {message}' if message else ''))
        self.bundle_id = bundle_id
        self.bundle_version = bundle_version


class NoRemoteAvailable(Exception):
    '''
    Thrown when we need a remote and we don't have one
    '''


class NotADescriptor(Exception):
    '''
    Thrown when a given file, string, or other object is offered as a descriptor, but does
    not represent a `Descriptor`
    '''


class DeployFailed(Exception):
    '''
    Thrown when bundle deployment fails for an apparently valid bundle
    '''


class NoAcceptableUploaders(DeployFailed):
    '''
    Thrown when, for all selected `Remotes <Remote>`, no `Uploaders <Uploader>` report
    that they can upload a given bundle
    '''
    def __init__(self, bundle_path):
        super(NoAcceptableUploaders, self).__init__(
                'Could not upload "%s" because no uploaders could handle it' %
                bundle_path)
        self.bundle_path = bundle_path



'''
Utilities for making objects that work with the `CLICommandWrapper`
'''
from .utils import FCN

DEFAULT_OWM_DIR = '.owm'


class IVar(object):
    '''
    A descriptor for instance variables amended to provide some attributes like
    default values, value types, etc.
    '''

    _instance_counter = 0

    def __init__(self, default_value=None, doc=None, value_type=str, name=None):
        self.name = ('_ivar_' + str(IVar._instance_counter)) if name is None else name
        IVar._instance_counter += 1
        self.default_value = default_value
        self.__doc__ = doc.strip() if doc is not None else doc
        self.value_type = value_type

    def __repr__(self):
        return '{}(name={}, doc={}, value_type={}, default_value={})'.format(
                FCN(type(self)), repr(self.name), repr(self.__doc__), repr(self.value_type), repr(self.default_value))

    __str__ = __repr__

    def __get__(self, target, typ=None):
        if target is None:
            return self
        return getattr(target, self.name, self.default_value)

    def __set__(self, target, value):
        setattr(target, self.name, value)

    @classmethod
    def property(cls, wrapped=None, *args, **kwargs):
        '''
        Creates a `PropertyIVar` from a method

        Typically, this will be used as a decorator for a method

        Parameters
        ----------
        wrapped : types.FunctionType or types.MethodType
            The function to wrap. optional: if omitted, returns a function that can be
            invoked later to create the `PropertyIVar`
        '''
        if wrapped is not None and \
                callable(wrapped) and \
                len(args) == 0 and \
                len(kwargs) == 0: # we only got the wrapped. Just execute now.
            return _property_update(wrapped)

        first_arg = wrapped
        aargs = list(args)

        def wrapper(wrapped):
            if first_arg is not None:
                args = [first_arg] + list(aargs)
            else:
                args = aargs
            return _property_update(wrapped, *args, **kwargs)
        return wrapper


def _property_update(wrapped, *args, **kwargs):
    res = PropertyIVar(*args, **kwargs)
    res.value_getter = wrapped
    wrapped_doc = getattr(wrapped, '__doc__', None)
    if wrapped_doc:
        res.__doc__ = wrapped_doc
    res.__doc__ = '' if res.__doc__ is None else res.__doc__.strip()
    return res


class SubCommand(object):
    '''
    A descriptor that wraps objects which function as sub-commands to
    `~owmeta_core.command.OWM` or to other sub-commands
    '''

    def __init__(self, cmd):
        self.cmd = cmd
        self.__doc__ = getattr(cmd, '__doc__', '')

    def __repr__(self):
        return '{}({})'.format(FCN(type(self)), repr(self.cmd))

    __str__ = __repr__

    def __get__(self, target, typ=None):
        if target is None:
            return self
        return self.cmd(target)


class PropertyIVar(IVar):
    '''
    An `.IVar` that functions similarly to a `property`


    Typically a PropertyIVar will be created by using `.IVar.property` as a decorator for
    a method like::

        class A(object):

            @IVar.property('default_value')
            def prop(self):
                return 'value'

    .. automethod:: __get__
    '''
    def __init__(self, *args, **kwargs):
        super(PropertyIVar, self).__init__(*args, **kwargs)
        self.value_getter = None
        self.value_setter = None
        self._setter_called_flag = '_' + self.name + '_is_set'

    def setter(self, fset):
        '''
        Decorator for the setter that goes along with this property.

        See Also
        --------
        property
        '''
        res = type(self)(default_value=self.default_value,
                         doc=self.__doc__,
                         value_type=self.value_type,
                         name=self.name)
        res.value_getter = self.value_getter
        res.value_setter = fset
        return res

    def __set__(self, target, value):
        if self.value_setter is None:
            raise AttributeError("can't set attribute")
        setattr(target, self._setter_called_flag, True)
        self.value_setter(target, value)

    def __get__(self, target, objecttype=None):
        ''' Executes the provided getter

        When the getter is first called, and when a setter is also defined, the setter will be called with the default
        value before the getter is called for the first time. _Even if the default_value is not set explicitly, the
        setter will still be called with 'None'.
        '''
        if target is None:
            return self

        if self.value_setter is not None and \
                not hasattr(target, self._setter_called_flag):
            setattr(target, self._setter_called_flag, True)
            self.value_setter(target, self.default_value)
        return self.value_getter(target)


class GeneratorWithData(object):
    def __init__(self, generator, header=None, text_format=None, default_columns=None, columns=None):
        self._gen = iter(generator)
        self.header = header
        if columns is tuple:
            if not self.header:
                raise Exception('Must provide a header if columns is `tuple`')
            columns = []
            for i, _ in enumerate(self.header):
                columns.append((lambda i: lambda x: x[i])(i))
            self.columns = columns
        else:
            self.columns = columns
        self.default_columns = default_columns
        self.text_format = text_format or None

    def __iter__(self):
        for m in self._gen:
            yield m

    def __next__(self):
        return next(self._gen)

    next = __next__


class GenericUserError(Exception):
    '''
    An error which should be reported to the user. Not necessarily an error that is
    the user's fault
    '''



from __future__ import absolute_import
from collections import namedtuple

from .utils import FCN


class Statement(namedtuple('Statement', ('subject', 'property', 'object', 'context'))):
    __slots__ = ()

    def contextualize(self, context):
        return Statement(self.subject, self.property, self.object, context)

    def to_quad(self):
        return (self.subject.idl,
                self.property.link,
                self.object.idl,
                self.context.identifier if self.context is not None else None)

    def to_triple(self):
        return (self.subject.idl, self.property.link, self.object.idl)

    def __repr__(self):
        return '{}(subj={}, prop={}, obj={}, context={})'.format(FCN(type(self)),
                                                                 repr(self.subject),
                                                                 repr(self.property),
                                                                 repr(self.object),
                                                                 repr(self.context))



'''
A collection of functions that produce `requests.Session` objects.

A few methods request a "session provider". The functions in here are providers of that kind
'''

import requests
from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache
from cachecontrol.heuristics import ExpiresAfter


def caching():
    '''
    Provides a `requests.Session` that puts cached responses in :file:`.owmeta_http_cache`

    In absence of explict cache-control headers, uses a heuristic of caching cacheable
    responses for up to a day.
    '''
    http_cache_directory = '.owmeta_http_cache'
    base_session = requests.Session()
    http_cache = FileCache(http_cache_directory)
    return CacheControl(base_session, cache=http_cache, heuristic=ExpiresAfter(days=1))



'''
Variables common to several `~owmeta_core.datasource.DataSource` and
`~owmeta_core.datasource.DataTranslator` implementations
'''
from rdflib.namespace import Namespace

from .. import BASE_SCHEMA_URL, BASE_DATA_URL

TRANS_NS = Namespace(BASE_SCHEMA_URL + '/translators/')
'''
Namespace for translators in owmeta-core. Not for use by packages downstream of
owmeta-core
'''

DS_NS = Namespace(BASE_SCHEMA_URL + '/data_sources/')
'''
Namespace for data sources in owmeta-core. Not for use by packages downstream of
owmeta-core
'''

DS_DATA_NS = Namespace(BASE_DATA_URL + '/data_sources/')
'''
Namespace for data sources in owmeta-core. Not for use by packages downstream of
owmeta-core
'''
